//*******************************************************
//
//	Color LCD Clock	Self Programmer
//
//	ATmega1284p
//		efuse:0xFD	BOD 2.7V
//		hfuse:0xD0	JTAG禁止, EEPROM消去保護
//		lfuse:0xE2	内蔵CR発振 8MHz
//
//	Compiler:	AVR-GCC		4.3.4
//				AVR-LIBC	1.6.7
//				binutils	2.20
//
//	2011.5.24	KAWAKAMI Yukio
//*******************************************************

#include	<avr/io.h>
#include	<avr/interrupt.h>
#include	<avr/pgmspace.h>
#include	<avr/eeprom.h>
#include	<avr/boot.h>
#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<ctype.h>

#include	"main.h"
#include	"lcd.h"

#include	<util/delay.h>

#define		FIRMWARE_FNAME	"CLCLOCK.BIN"

volatile uchar	Timer;

extern int CurX, CurY;

#include	"pff.h"

FATFS	FatFs;

const char	*Fname = FIRMWARE_FNAME;

//void disk_timerproc(void);	// diskio.c
extern volatile
BYTE Timer1, Timer2;	/* 100Hz decrement timer */


//*******************************************************
//	タイマー０比較一致Ａ割り込み
//	100Hz
//*******************************************************
ISR(TIMER0_COMPA_vect){

	if (Timer)	--Timer;

	// diskio.c
	if (Timer1)	--Timer1;
	if (Timer2)	--Timer2;
}


//*******************************************************
//	割り込み停止
//*******************************************************
void stop_interrupt(void){
	uchar	r;

	cli();
	WDR;

	// 割り込み禁止
	UCSR0B = 0;
	TIMSK0 = 0;
	EIMSK = 0;

	// 割り込みフラグクリア
	UCSR0A = (1<<TXC0);
	r = UDR0;
	TIFR0 = 0xFF;
	EIFR |= (1<<INTF1);
}


//*******************************************************
//	メインルーチン
//*******************************************************
int main(void){
	uchar	r;

	cli();

	if (SMCR == 0){
		// 電源ONリセット
		boot();
	}

	PORTA = (1<<RSTB)|(1<<LWR)|(1<<LRD)|(1<<LCS);
	DDRA = (1<<RSTB)|(1<<LWR)|(1<<LRD)|(1<<LRES)|(1<<LRS)|(1<<LCS);

	PORTB = (1<<MOSI)|(1<<SDCS)|(1<<MCS);
	DDRB = (1<<SCK)|(1<<MOSI)|(1<<SDCS)|(1<<RDATA)|(1<<RCLK)|(1<<CCE)|(1<<MCS);

	PORTC = 0;
	DDRC = 0xFF;

	PORTD = (1<<PD7)|(1<<TXD)|(1<<RXD);
	DDRD = (1<<BCE)|(1<<BR2)|(1<<BR1)|(1<<RWR);

	if (MCUSR & (1<<WDRF)){
		// ウォッチドッグリセット
	}
	MCUSR = 0;

	WDR;
	WDTCSR = (1<<WDCE)|(1<<WDE);	// WDCEとWDEを1にしてタイマー変更をする
	WDTCSR = 0;						// ウォッチドッグ停止

	stop_interrupt();	// 割り込み停止

	// ブートローダ領域に割り込みベクタ移動
	r = MCUCR;
	MCUCR = r | (1<<IVCE);
	MCUCR = r | (1<<IVSEL);

	TCCR0A = (1<<WGM01);			// CTC動作
	TCCR0B = (1<<CS02)|(1<<CS00);	// プリスケーラ 1/1024
	OCR0A = F_CPU/1024/100 - 1;		// タイマカウント 100Hz
	TIMSK0 = (1<<OCIE0A);			// 比較一致Ａ割り込み許可

	SMCR = (1<<SE);			// スリープ許可 アイドルモード

	// TWI, ADC 電力カット
	PRR0 = (1<<PRTWI)|(1<<PRADC);
	ACSR = (1<<ACD);			// アナログコンパレータ禁止

	sei();

	wait_timer(10);

	LCD_init();

	PORTD |= (1<<BCE);	// バックライト ON
	PORTD |= (1<<BR2);	//

	pf_mount(&FatFs);

	if (pf_open(Fname) != FR_OK){
		print_P(PSTR("Open Err"));
	} else {
		ushort	code;
		WORD	n = 0;
		pf_read((void *)&code, 2, &n);	// 先頭2bytesチェック
		if ((FatFs.fsize < 8000)||
			(FatFs.fsize > 0x1E000)||
			((code & 0xFE0E) != 0x940C)){
			print_P(PSTR("File Err"));
		} else {
			uchar	i;
			ushort	fdate, ftime;
			eeprom_write_byte(EEPROM_SELF_PROGRAM, 'W');	// 書き換え中
			eeprom_busy_wait();
			for(i = 0; i < 3; i++){
				print_P(PSTR("Writing\n"));
				pf_open(Fname);
				update_firm();
				print_P(PSTR("Verify\n"));
				pf_open(Fname);
				if (verify_firm()){
					eeprom_write_byte(EEPROM_SELF_PROGRAM, 0xFF);	// 書き換え終了
					get_fdatetime(&fdate, &ftime);
					eeprom_write_word(EEPROM_FIRM_FDATE, fdate);
					eeprom_write_word(EEPROM_FIRM_FTIME, ftime);
					eeprom_busy_wait();
					break;
				}
			}
		}
	}

	wait_timer(250);

	PORTD &= ~(1<<BR2);
	PORTD &= ~(1<<BR1);

	boot();

	for(;;){
		WDR;
		SLEEP;
	}
}


//****************************************************************
//	ファイルの更新時間取得
//****************************************************************
void get_fdatetime(ushort *fdate, ushort *ftime){
	FRESULT	res;
	FILINFO	finfo;
	DIR		dir;
	char	path[16];

	path[0] = 0;
	res = pf_opendir(&dir, path);
	if (res != FR_OK)	return;
	for(;;){
		res = pf_readdir(&dir, &finfo);
		if (res != FR_OK || finfo.fname[0] == 0)	break;
		if (!(finfo.fattrib & AM_DIR)){
			if (strcmp(finfo.fname, Fname) == 0){
				*fdate = finfo.fdate;
				*ftime = finfo.ftime;
				return;
			}
		}
	}
}


//*******************************************************
//	自己フラッシュROM書き換え
//	ページ単位で書き換えることに注意
//	ATmega1284pは 1ページ 256bytes
//	メインプログラムから呼び出すので 1FF40H固定アドレス
//	また、この呼び出しの前後で cli(); sei(); をしておくこと
//*******************************************************
ADRS_1FF40H void write_flash(ulong adrs, ushort *buff){
	uchar	i;
	ulong	page;

	page = adrs;
	boot_page_erase_safe(page);
	for (i = 0; i < SPM_PAGESIZE / 2; i++){
		boot_page_fill_safe(adrs, *buff++);
		adrs += 2;
	}
	boot_page_write_safe(page);
	boot_rww_enable_safe();
}


//****************************************************************
//	セルフプログラム
//****************************************************************
void update_firm(void){
	uint	i;
	ushort	w;
	ushort	sum = 0;
	ulong	page = 0;
	uchar	buff[512];
	WORD	n = 0;
	char	msg[20];
	FRESULT	res = 0;

	while(page < FatFs.fsize){
		WDR;

		// MMC/SDから１セクタ分読み込み
		memset(buff, 0xFF, 512);
		if ((res = pf_read(buff, 512, &n)) != FR_OK){
			break;
		}

		for (i = 0; i < 512; i += 2){
			if ((i & (SPM_PAGESIZE - 1)) == 0){
				// ページ消去
				boot_page_erase_safe(page);
				sum = 0xFFFF;
			}
			w = *(ushort *)(&buff[i]);
			sum &= w;
			cli();
			boot_page_fill_safe(page + i, w);
			sei();
			if ((i & (SPM_PAGESIZE - 1)) == (SPM_PAGESIZE - 2)){
				if (sum != 0xFFFF){	// 全て FFFF なら書き込まない
					// ページ書き込み
					boot_page_write_safe(page);
				}
				page += SPM_PAGESIZE;
			}
		}

		// RWWセクションを有効にする
		boot_rww_enable_safe();

		// 液晶表示
		ltoa(FatFs.fptr, msg, 16);
		strupr(msg);
		print(msg);
		putch(CR);
	}

	// RWWセクションを有効にする
	boot_rww_enable_safe();

	if (res == FR_OK){
		print_P(PSTR("\nOK\n"));
	} else {
		print_P(PSTR("\nRead Err\n"));
	}
}


//****************************************************************
//	Verify
//****************************************************************
char verify_firm(void){
	ulong	adrs;
	ushort	data = 0;
	ushort	rom;
	WORD	n;
	char	msg[8];

	adrs = 0;
	while(adrs < FatFs.fsize){
		WDR;

		// 液晶表示
		if ((adrs & 0x00FF) == 0){
			ltoa(adrs, msg, 16);
			strupr(msg);
			print(msg);
			putch(CR);
		}

		if (pf_read((void *)&data, 2, &n) != FR_OK){
			print_P(PSTR("\nRead Err\n"));
			return 1;
		}

		rom = pgm_read_word_far(adrs);
		if (rom != data){
			ltoa(adrs, msg, 16);
			strupr(msg);
			print(msg);
			putch(':');
			ltoa(data, msg, 16);
			strupr(msg);
			print(msg);
			print_P(PSTR("->"));
			ltoa(rom, msg, 16);
			strupr(msg);
			print(msg);
			print_P(PSTR(" Verify Err\n"));
			return 0;
		}

		adrs += 2;
	}
	ltoa(adrs, msg, 16);
	strupr(msg);
	print(msg);
	putch(CR);

	print_P(PSTR("\nOK\n"));
	return 1;
}


//****************************************************************
//	０番地からのプログラム実行
//****************************************************************
void boot(void){
	ushort	code;

	// 書き換え中ならブートしない
	if (eeprom_read_byte((uint8_t *)EEPROM_SELF_PROGRAM) != 0xFF){
		return;
	}

	// コードチェック。最初の jmp命令を検査
	code = pgm_read_word_far(0);
	if ((code & 0xFE0E) == 0x940C){
		uchar	r;

		stop_interrupt();	// 割り込み停止

		// 割り込みベクタを戻す
		r = MCUCR;
		MCUCR = r | (1<<IVCE);
		MCUCR = r & ~(1<<IVSEL);

		// 0番地にジャンプ(RESET)
		asm volatile("jmp	0\n"::);
	}
}


//*******************************************************
//	時間待ち
//*******************************************************
void wait_timer(uchar t){

	Timer = t;
	while(Timer){
		WDR;
		SLEEP;
	}
}

